/**
 * GPU Programming with CUDA
 *
 * Demo code 1: device properties and error checking
 *
 * Compilation, assuming envvars properly set
 *
 *   g++ -I$CUDA_INC_PATH -DDEBUG -Wall 01-deviceinfo.cpp -L$CUDA_LIB_PATH -lcudart
 * or
 *   nvcc 01-deviceinfo.cpp -Xcompiler=-Wall
 *
 * (c) 2014-16 dominik.goeddeke@ians.uni-stuttgart.de
 */


// C++ includes
#include <iostream>

// CUDA include, only necessary if not compiling with nvcc
#include <cuda_runtime.h>


int main (int argc, char **argv)
{
  //
  // get number of devices in the system, and perform proper error checking:
  // all CUDA API functions return a cudaError_t "object"
  //
  int numDevices(-1);
  cudaError_t status = cudaGetDeviceCount (&numDevices);
  if (status != cudaSuccess)
  {
    std::cerr << "CUDA ERROR (cudaGetDeviceCount): " << cudaGetErrorString(status) << ". Exiting..." << std::endl;
    return 1;
  }
  std::cout << "Number of devices in this machine: " << numDevices << std::endl;

  //
  // manually treat the case that this is run on a system without a CUDA-capable
  // GPU that has the toolkit installed (like our servers, which share the software
  // stack even though not all of them have a GPU)
  //
  if (numDevices == 0)
  {
    std::cerr << "No CUDA-capable device found." << std::endl;
    return 2;
  }

  //
  // loop over all devices and query (some of) their properties
  //
  for (int device(0); device<numDevices; ++device)
  {
    // set CUDA device
    status = cudaSetDevice (device);
    if (status != cudaSuccess)
    {
      std::cerr << "CUDA ERROR (cudaSetDevice): " << cudaGetErrorString(status) << ". Exiting..." << std::endl;
      return 3;
    }
    // get device properties
    cudaDeviceProp prop;
    status = cudaGetDeviceProperties (&prop, device);
    if (status != cudaSuccess)
    {
      std::cerr << "CUDA ERROR (cudaGetDeviceProperties): " << cudaGetErrorString(status) << ". Exiting..." << std::endl;
      return 4;
    }
    // print out device name and compute capabilities
    std::cout << "Device " << device << ": " << prop.name;
    std::cout << " (cc " << prop.major << "." << prop.minor << ")" << std::endl;
    std::cout << "Important: compile all future codes with \"nvcc -arch=sm_";
    std::cout << prop.major << prop.minor << "\" on this machine." << std::endl;
  }
  
  return 0;
}

