/**
 * GPU Programming with CUDA
 *
 * Demo code 2: vector addition
 *              memory management, simple kernels, proper error checking
 *
 * Make sure to use correct argument to cudaSetDevice() around line 78
 *
 * Compilation (use appropriate -arch=smXY as recommended by demo code 01-deviceinfo.cpp):
 *
 *   nvcc -arch=sm_52 02-vecadd.cu -Xcompiler=-Wall
 *
 * (c) 2014-16 dominik.goeddeke@ians.uni-stuttgart.de
 */


/****************************************************************************
 * includes
 ****************************************************************************/
#include <iostream>        // for std::cout and friends 
#include <cstdlib>         // for exit()
#include <cuda_runtime.h>  // for CUDA, not technically necessary when compiling through nvcc


/*****************************************************************************
 * a small convenience function to reduce redundancies, essentially the same
 * as the error checking code in the last example, but outsourced into a 
 * standalone routine and equipped with labels to have meaningful and 
 * pretty-printed error messages
 *****************************************************************************/
void checkForErrors (const cudaError_t status, const char *label, const int line, const char *file)
{
  if (status != cudaSuccess)
  {
    std::cerr << "CUDA ERROR (" << label << ") ";
    std::cerr << "at line " << line << " of file " << file << ":" << std::endl;
    std::cerr << cudaGetErrorString(status) << ". Exiting..." << std::endl;
    exit(1);
  }
}


/*****************************************************************************
 * our kernel, renamed from mykernel (slides) to vecadd, otherwise unchanged
 *****************************************************************************/
__global__ void vecadd_gpu (int *c, const int *a, const int *b, const int N)
{
  const int i(threadIdx.x + blockDim.x*blockIdx.x);
  if (i<N)
    c[i] = a[i] + b[i];
}


/*****************************************************************************
 * ... and for didactical reasons, the same for the CPU
 ****************************************************************************/
void vecadd_cpu (int *c, const int *a, const int *b, const int N)
{
  for (int i(0); i<N; ++i)
    c[i] = a[i] + b[i];
}


/*****************************************************************************
 * main
 *****************************************************************************/
int main (int argc, char **argv)
{
  //
  // get number of devices in the system, same as in the previous demo code
  //
  int numDevices(-1);
  cudaError_t status = cudaGetDeviceCount (&numDevices);
  checkForErrors (status, "cudaGetDeviceCount", __LINE__, __FILE__);
  std::cout << "Number of devices in this machine: " << numDevices << std::endl;
  if (numDevices == 0)
  {
    std::cerr << "No CUDA-capable device found." << std::endl;
    return 2;
  }

  //
  // select one device and print out basic properties (unchanged)
  //
  const int myDevice(0);
  status = cudaSetDevice (myDevice);
  checkForErrors (status, "cudaSetDevice", __LINE__, __FILE__);
  cudaDeviceProp prop;
  status = cudaGetDeviceProperties (&prop, myDevice);
  checkForErrors (status, "cudaGetDeviceProperties", __LINE__, __FILE__);
  std::cout << "Device " << myDevice << ": " << prop.name;
  std::cout << " (cc " << prop.major << "." << prop.minor << ")" << std::endl;

  //
  // define problem size and allocate arrays on host and device
  //
  const int N(1234567);
  int *a_cpu = new int[N];
  int *b_cpu = new int[N];
  int *c_cpu = new int[N];
  int *a_gpu, *b_gpu, *c_gpu;
  status = cudaMalloc((void**)&a_gpu, N*sizeof(int));
  checkForErrors (status, "cudaMalloc a_gpu", __LINE__, __FILE__);
  status = cudaMalloc((void**)&b_gpu, N*sizeof(int));
  checkForErrors (status, "cudaMalloc b_gpu", __LINE__, __FILE__);
  status = cudaMalloc((void**)&c_gpu, N*sizeof(int));
  checkForErrors (status, "cudaMalloc c_gpu", __LINE__, __FILE__);

  //
  // fill CPU arrays with some arbitrary data
  //
  for (int i(0); i<N; ++i)
  {
    a_cpu[i] = i+1;
    b_cpu[i] = 10*i;
  }
  
  //
  // copy data to the GPU
  //
  status = cudaMemcpy (a_gpu, a_cpu, N*sizeof(int), cudaMemcpyHostToDevice);
  checkForErrors (status, "cudaMemcpy a H2D", __LINE__, __FILE__);
  status = cudaMemcpy (b_gpu, b_cpu, N*sizeof(int), cudaMemcpyHostToDevice);
  checkForErrors (status, "cudaMemcpy b H2D", __LINE__, __FILE__);
  
  //
  // perform computation on CPU, outsourced into a "kernel" for didactical reasons
  //
  vecadd_cpu(c_cpu, a_cpu, b_cpu, N);
  
  //
  // launch vector addition kernel on the GPU
  //
  dim3 block(256,1,1);
  dim3 grid( (unsigned)ceil(N/(double)block.x), 1, 1);
  std::cout << "Launching " << grid.x << " blocks with ";
  std::cout << block.x << " threads each." << std::endl;
  vecadd_gpu<<<grid,block>>>(c_gpu, a_gpu, b_gpu, N);
  cudaDeviceSynchronize ();  // explained later
  checkForErrors (cudaGetLastError(), "kernel", __LINE__, __FILE__);

  //
  // copy result back from GPU to CPU
  // (we no longer need a_cpu, so we reuse this array)
  //
  status = cudaMemcpy(a_cpu, c_gpu, N*sizeof(int), cudaMemcpyDeviceToHost);
  checkForErrors(status, "cudaMemcpyDeviceToHost", __LINE__, __FILE__);

  //
  // compare results
  //
  bool success(true);
  for (int i(0); i<N; ++i)
  {
    if (a_cpu[i] != c_cpu[i])
    {
      success = false;
      std::cout << "Mismatch at index " << i << std::endl;
    }
  }
  if (success)
  {
    std::cout << "Result comparison successful." << std::endl;
  }

  //
  // clean up
  //
  delete [] a_cpu;
  delete [] b_cpu;
  delete [] c_cpu;
  status = cudaFree(a_gpu);
  checkForErrors(status, "cudaFree(a_gpu)", __LINE__, __FILE__);
  status = cudaFree(b_gpu);
  checkForErrors(status, "cudaFree(b_gpu)", __LINE__, __FILE__);
  status = cudaFree(c_gpu);
  checkForErrors(status, "cudaFree(c_gpu)", __LINE__, __FILE__);
  
  cudaDeviceReset();

  return 0;
}

