/**
 * GPU Programming with CUDA
 *
 * Demo code 3: blocks and threads
 *
 * Make sure to use correct argument to cudaSetDevice() 
 *
 * Compilation (use appropriate -arch=smXY as recommended by demo code 01-deviceinfo.cpp)
 *
 *   nvcc -arch=sm_52 03-blocks-threads.cu -Xcompiler=-Wall
 *
 * (c) 2014-16 dominik.goeddeke@ians.uni-stuttgart.de
 */


/*****************************************************************************
 * includes
 *****************************************************************************/
#include <iostream>        // for std::cout and friends 
#include <iomanip>         // for pretty-printing
#include <cstdlib>         // for exit()
#include <cuda_runtime.h>  // for CUDA, not technically necessary when compiling
                           // through nvcc


/*****************************************************************************
 * convenience pretty printer
 *****************************************************************************/
void checkForErrors (const cudaError_t status, const char *label, const int line, const char *file)
{
  if (status != cudaSuccess)
  {
    std::cerr << "CUDA ERROR (" << label << ") ";
    std::cerr << "at line " << line << " of file " << file << ":" << std::endl;
    std::cerr << cudaGetErrorString(status) << ". Exiting..." << std::endl;
    exit(1);
  }
}


/*****************************************************************************
 * kernels for assignment 1
 *****************************************************************************/

// TODO



/*****************************************************************************
 * main
 *****************************************************************************/
int main (int argc, char **argv)
{
  //
  // get number of devices in the system
  //
  int numDevices(-1);
  cudaError_t status = cudaGetDeviceCount (&numDevices);
  checkForErrors (status, "cudaGetDeviceCount", __LINE__, __FILE__);
  std::cout << "Number of devices in this machine: " << numDevices << std::endl;
  if (numDevices == 0)
  {
    std::cerr << "No CUDA-capable device found." << std::endl;
    return 2;
  }

  //
  // select one device and print out basic properties
  //
  const int myDevice(0);
  status = cudaSetDevice (myDevice);
  checkForErrors(status, "cudaSetDevice", __LINE__, __FILE__);
  cudaDeviceProp prop;
  status = cudaGetDeviceProperties (&prop, myDevice);
  checkForErrors(status, "cudaGetDeviceProperties", __LINE__, __FILE__);
  
  std::cout << "Device name       : " << prop.name << std::endl;
  std::cout << "Compute capability: " << prop.major << "." << prop.minor << std::endl;
  std::cout << "Warp size                     : " << prop.warpSize << std::endl;
  std::cout << "Max threads per core          : " << prop.maxThreadsPerMultiProcessor << std::endl;
  std::cout << "Max threads per block         : " << prop.maxThreadsPerBlock << std::endl;
  std::cout << "Max block dimensions          : " << prop.maxThreadsDim[0] << "x" << prop.maxThreadsDim[1] << "x"<< prop.maxThreadsDim[2] << std::endl;
  std::cout << "Max grid dimensions           : " << prop.maxGridSize[0] << "x" << prop.maxGridSize[1] << "x"<< prop.maxGridSize[2] << std::endl;
  std::cout << std::endl;

  //
  // assignment 1.1: memory roundtrip
  //
  const int N(12);
  int* array1_cpu = new int[N];
  for (int i(0); i<N; ++i)
  	array1_cpu[i] = -1;

  // TODO allocate and memcpy

  int* array2_cpu = new int[N];
  // TODO: memcpy GPU back into array2_cpu
  for (int i(0); i<N; ++i)
  {
    if (array1_cpu[i] != array2_cpu[i])
      std::cout << "Mismatch at index " << i << std::endl;
  }
  std::cout << "Memory roundtrip finished" << std::endl;
  std::cout << std::endl;

  //
  // assignment 1.2a
  //
  std::cout << "Assignment 1.2 (blockDim), 3B4T" << std::endl;
  
  std::cout << "Assignment 1.2 (blockIdx), 3B4T" << std::endl;
  
  std::cout << "Assignment 1.2 (threadIdx), 3B4T" << std::endl;
  
  //
  // assignment 1.2a
  //
  std::cout << "Assignment 1.2 (blockDim), 4B3T" << std::endl;
  
  std::cout << "Assignment 1.2 (blockIdx), 4B3T" << std::endl;
  
  std::cout << "Assignment 1.2 (threadIdx), 4B3T" << std::endl;
  
  //
  // cleanup
  //
  delete [] array1_cpu;
  delete [] array2_cpu;
  
  return 0;
}

